package str.classtree.atermify;

import org.apache.bcel.Constants;
import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import aterm.*;
import static str.classtree.atermify.ATermifyUtil.*;
import str.aterm.ATermifier;
import str.bcel.EnclosingMethod;
import static str.aterm.ATermUtil.*;

/**
 * Standard class attributes:
 *
 * InnerClasses, EnclodingMethod, Synthetic, SourceFile, Signature,
 * Deprecated
 */
public class ClassAttributeATermifier extends AbstractAttributeATermifier
{
  private JavaClass _class;

  public ClassAttributeATermifier(JavaClass clazz)
  {
    super();
    _class = clazz;
  }

  public ATerm convert(Attribute attr)
  {
    ATerm result;

    if(attr instanceof SourceFile)
    {
      result = appl("SourceFile", astring(((SourceFile)attr).getSourceFileName()));
    }
    else if(attr instanceof Synthetic)
    {
      result = buildSyntheticAttribute((Synthetic) attr);
    }
    else if(attr instanceof InnerClasses)
    {
      result = buildInnerClassesAttribute((InnerClasses) attr);
    }
    else if(attr instanceof org.apache.bcel.classfile.Deprecated)
    {
      result = buildDeprecatedAttribute((org.apache.bcel.classfile.Deprecated) attr);
    }
    else if(attr instanceof EnclosingMethod)
    {
      EnclosingMethod m = (EnclosingMethod) attr;

      ATerm method;
      if(m.getMethodIndex() == 0)
      {
	method = appl("None");
      }
      else
      {
	method = appl("Some",
	  appl("Method",
	    appl("Name", astring(m.getMethodName())), 
	    buildMethodDescriptor(m.getReturnType(), m.getArgumentTypes())
	  ));
      }

      result = appl("EnclosingMethod", appl("Class", astring(m.getClassName())), method);
    }
    else if(attr instanceof Signature)
    {
      result = appl("ClassSignature", astring(((Signature) attr).getSignature()));
    }
    else
    {
      result = buildUnknownAttribute(attr);
    }

    return result;
  }

  private ATerm buildInnerClassesAttribute(InnerClasses attr)
  {
    ATermList result = _factory.makeList();
    InnerClass[] inners = attr.getInnerClasses();

    for(int i = inners.length - 1; i >= 0; i--)
    {
      InnerClass inner = inners[i];

      ATerm inner_class;
      ATerm outer_class;
      ATerm inner_name;

      int inner_index = inner.getInnerClassIndex();
      if(inner_index == 0)
	inner_class = appl("None");
      else 
	inner_class = appl("Some", astring(getClassName(inner_index)));

      int outer_index = inner.getOuterClassIndex();
      if(outer_index == 0)
	outer_class = appl("None");
      else 
	outer_class = appl("Some", astring(getClassName(outer_index)));

      int name_index = inner.getInnerNameIndex();
      if(name_index == 0)
	inner_name = appl("None");
      else 
	inner_name = appl("Some", astring(getConstantUTF8(name_index)));
     
      result = result.insert(
	appl("InnerClass",
	  appl("InnerClass", inner_class),
	  appl("OuterClass", outer_class),
	  appl("InnerName", inner_name), 
	  appl("AccessFlags", buildClassFlags(inner.getInnerAccessFlags()))));
    }

    return appl("InnerClasses", result);
  }

  private final String getClassName(int index)
  {
    ConstantPool cp = _class.getConstantPool();
    String result =
      cp.getConstantString(index, Constants.CONSTANT_Class);
    result = Utility.compactClassName(result, false);

    return result;
  }

  private final String getConstantUTF8(int index)
  {
    ConstantPool cp = _class.getConstantPool();
    ConstantUtf8 c = (ConstantUtf8) cp.getConstant(index, Constants.CONSTANT_Utf8);
    return c.getBytes();
  }
}
